package com.danitheskunk.skunkworks.backends.gl;

import com.danitheskunk.skunkworks.WindowStretchMode;

import static org.lwjgl.opengl.GL11.*;
import static org.lwjgl.opengl.GL20.glVertexAttrib2f;
import static org.lwjgl.opengl.GL41.glProgramUniform2f;

public class Scaler extends Filter {
	private static final String fragmentSource = """
		#version 450
		layout(location = 1) in vec2 texCoord;
		layout(binding = 0) uniform sampler2D tex;
		out vec4 color;
		void main() {
			//color = vec4(vec2(texCoord).x/1000.f, 0.2f, 0.7f, 1.0f);
			vec4 col = texture(tex, texCoord);
			if(col.a > 0.0) {
				color = col;
			}
			//color = texelFetch(tex, ivec2(texCoord), 0);
			//color = vec4(1.0, 1.0, 1.0, 1.0);
			//color = vec4(texCoord.xy, 0.0, 1.0);
		}
		""";
	private static final String vertexSource = """
		#version 450
		layout(location = 0) in vec2 pos;
		layout(location = 1) in vec2 texCoord;
		layout(location = 2) uniform vec2 windowSize;
		layout(location = 1) out vec2 out_texCoord;
		void main() {
			gl_Position = vec4(pos / windowSize * vec2(2.0f, 2.0f) + vec2(-1.0f, -1.0f), 0.0f, 1.0f);
			out_texCoord = texCoord;
		}
		""";

	private final Program program;
	private WindowStretchMode stretchMode;

	public Scaler(Framebuffer input) {
		super(input, true);
		program = new Program(vertexSource, fragmentSource);
	}

	@Override
	protected void process() {
		var outputSize = output.getSize();
		var inputSize = input.getSize();
		glViewport(0, 0, outputSize.getX(), outputSize.getY());

		program.use();
		glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

		glProgramUniform2f(
			program.getProgram(),
			program.getUniformLocation("windowSize"),
			outputSize.getX(),
			outputSize.getY()
		);

		int tlx1 = 0;
		int tly1 = 0;
		int tlx2 = 0;
		int tly2 = 0;

		switch(stretchMode) {
			case STRETCH -> {
				tlx2 = outputSize.getX();
				tly2 = outputSize.getY();
			}
			case ASPECT -> {
				float scalex = (float) outputSize.getX() /
					(float) inputSize.getX();
				float scaley = (float) outputSize.getY() /
					(float) inputSize.getY();
				float scale = Math.min(scalex, scaley);

				int xoff = (int) (
					(outputSize.getX() - inputSize.getX() * scale) / 2
				);
				int yoff = (int) (
					(outputSize.getY() - inputSize.getY() * scale) / 2
				);

				tlx1 = xoff;
				tly1 = yoff;
				tlx2 = (int) (xoff + inputSize.getX() * scale);
				tly2 = (int) (yoff + inputSize.getY() * scale);
			}
			case INTEGER -> {
				int scalex = outputSize.getX() / inputSize.getX();
				int scaley = outputSize.getY() / inputSize.getY();
				int scale = Math.max(1, Math.min(scalex, scaley));

				int xoff = (outputSize.getX() - inputSize.getX() * scale) / 2;
				int yoff = (outputSize.getY() - inputSize.getY() * scale) / 2;

				tlx1 = xoff;
				tly1 = yoff;
				tlx2 = xoff + inputSize.getX() * scale;
				tly2 = yoff + inputSize.getY() * scale;
			}
		}


		var texCoordIndex = program.getAttribLocation("texCoord");
		glBegin(GL_TRIANGLES);
		//counterclockwise triangles
		glVertexAttrib2f(texCoordIndex, 0, 1);
		glVertex2i(tlx1, tly2);
		glVertexAttrib2f(texCoordIndex, 1, 0);
		glVertex2i(tlx2, tly1);
		glVertexAttrib2f(texCoordIndex, 0, 0);
		glVertex2i(tlx1, tly1);

		glVertexAttrib2f(texCoordIndex, 1, 0);
		glVertex2i(tlx2, tly1);
		glVertexAttrib2f(texCoordIndex, 0, 1);
		glVertex2i(tlx1, tly2);
		glVertexAttrib2f(texCoordIndex, 1, 1);
		glVertex2i(tlx2, tly2);

		glEnd();

	}

	public void setStretchMode(WindowStretchMode stretchMode) {
		this.stretchMode = stretchMode;
	}
}
